//
// Created by mockie on 30/01/2021.
//

#define DEBUG

#include <iostream>
#include <server/Game.hpp>
#include "server/ServerManager.hpp"
#include "network/TCPServer.hpp"
#include "Exceptions.hpp"

std::queue<uint16_t> Client::freeId = std::queue<uint16_t>();

const std::map<unsigned short, void (ServerManager::*)(Client &client, const std::string &message)> ServerManager::eventsHandlers = {
        {request::START_GAME, &ServerManager::startGame}
};

void ServerManager::run()
{
    _socketUDP->read();
    _server->acceptClient();
    _server->run();
    _context.run();
    std::cout << "hello the server is multithreaded oh wow !" << std::endl;
}

std::string ServerManager::prepareHeader(unsigned short code, unsigned int size)
{
    std::string response;
    network::TCPHeader header = {code, size};

    response.reserve(size + sizeof(TCPHeader));
    response.append(reinterpret_cast<char *>(&header), sizeof(TCPHeader));

    return response;
}

bool ServerManager::handleAccept(std::shared_ptr<network::ISocket> &connection)
{
    std::cout << "{SERVER} New connection" << std::endl;
    std::shared_ptr<Client> client = std::make_shared<Client>(connection);

    client->port = _port;
    _clients.emplace_back(client);

    auto ptr = client.get();

    connection->setWriteFct([this, ptr] () {this->writeHandler(*ptr);});
    connection->setReadFct([this, ptr] (const std::string &str) {return this->readHandler(*ptr, str);});
    connection->setDisconnectFct([this, ptr] () {this->disconnectHandler(*ptr);});

    std::string response = prepareHeader(network::response::OK, sizeof(unsigned short));

    response.append(reinterpret_cast<char *>(&(client->id)), sizeof(unsigned short));
    connection->write(response);

    return true;
}

void ServerManager::writeHandler(Client &client)
{
    if (!client.isReading) {
        client.socket->read();
        client.isReading = true;
    }
}

int ServerManager::readHandler(Client &client, const std::string &str)
{
    std::cout << "{SERVER} Enter to readHandler" << std::endl;
    network::TCPHeader header = {0, 0};

    std::cout << "Str length :" << str.length() << std::endl;
    if (str.length() < sizeof(TCPHeader)) {
        client.socket->read();
        client.isReading = true;
        return 0;
    }
    std::memcpy(&header, str.c_str(), sizeof(TCPHeader));
    if (str.length() < sizeof(TCPHeader) + header.messageSize) {
        client.socket->read();
        client.isReading = true;
        return 0;
    }

    client.isReading = false;
    std::string message;
    message.reserve(header.messageSize);
    message.append(str.c_str() + sizeof(TCPHeader), header.messageSize);

    auto fct = eventsHandlers.find(header.nb);
    if (fct != eventsHandlers.end())
        (this->*fct->second)(client, message);
    else
        this->unknownRequest(client, message);
    return sizeof(TCPHeader) + header.messageSize;
}

void ServerManager::disconnectHandler(Client &client)
{
    if (client.idRoom != -1) {
        if (!_games[client.idRoom]->removePlayer(client)) {
            std::cout << "Dont delete client" << std::endl;
            client.connected = false;
            client.idRoom = -1;
            client.socket = nullptr;
            return;
        } else {
            if (_games[client.idRoom]->getPlayers().empty()) {
                _games[client.idRoom].reset();
                _games[client.idRoom] = nullptr;
            }
        }
    }
    auto it = std::find_if(
            _clients.begin(), _clients.end(),
            [&client] (const std::shared_ptr<Client> &elem) {
                return &client == elem.get();
            });
    if (it == _clients.end())
        return;
    _clients.erase(it);
}


[[maybe_unused]] void ServerManager::startGame(Client &client, const std::string &)
{
    std::string response;

    if (client.idRoom == -1) {
        std::cout << "{SERVER} Start game bad request : id room == -1" << std::endl;
        response = prepareHeader(response::BAD_REQUEST, 0);
        client.socket->write(response);
        return;
    }
    if (client.idRoom > static_cast<int16_t>(_games.size()) || _games[client.idRoom] == nullptr) {
        std::cout << "{SERVER} Start game bad request : game not found" << std::endl;
        response = prepareHeader(response::NOT_FOUND, 0);
        client.socket->write(response);
        return;
    }
    if (_games[client.idRoom]->isStart()) {
        std::cout << "{SERVER} Start game bad request : game is already start" << std::endl;
        response = prepareHeader(response::BAD_REQUEST, 0);
        client.socket->write(response);
    } else {
        response = prepareHeader(response::GAME_STARTING, 0);
        client.socket->write(response);
    }
}





void ServerManager::unknownRequest(Client &client, const std::string &)
{
    std::cout << "Unkown request" << std::endl;
    std::string response = prepareHeader(response::NOT_FOUND, 0);
    client.socket->write(response);
}








int getPort(int argc, char **argv) {
    int port = 0;

    if (argc < 2) {
        port = 8080;
        std::cout << "No port specified as argument, defaults to 8080." << std::endl;
    } else {
        for (auto letter : std::string(argv[1])) {
            if (!std::isdigit(letter)) {
                throw client::ServerException(std::string(argv[1]) + " : Invalid port.");
            }
        }

        port = std::stoi(argv[1]);
        if (port < 0 || port >= 65535)
            throw client::ServerException(std::string(argv[1]) + ": Invalid port. Port must be in the 0-65535 range.");
    }
    std::cout << "The server will be started on port " << port << std::endl;

    return port;
}

void ServerManager::initialize() {
    _server = std::make_unique<network::TCPServer>(_context, _port);
    _server->setAcceptHandler([this] (std::shared_ptr<network::ISocket> &connection) {
        return this->handleAccept(connection);
    });
    _socketUDP = network::UDPSocket::create(_context, _port);
    _socketUDP->setReadFct([this](const std::string &msg) -> int {
#ifdef DEBUG
        std::cout << "UDP size " << msg.size() << std::endl;
#endif
        UDPEventClient eventClient;

        std::memcpy(&eventClient, msg.c_str(), sizeof(UDPEventClient));
        if (eventClient.messageSize + sizeof(UDPEventClient) != msg.size())
            return (msg.size());
        if (eventClient.idRoom < 0 || static_cast<int16_t>(_games.size()) < eventClient.idRoom || _games[eventClient.idRoom] == nullptr)
            return (msg.size());
        _games[eventClient.idRoom]->addClientData(eventClient.idClient, msg.data() + sizeof(UDPEventClient));
        return (msg.size());
    });
}

ServerManager::ServerManager(int argc, char **argv) {
    _port = getPort(argc, argv);
}
