/*
** EPITECH PROJECT, 2020
** R-Type
** File description:
** UDPSocket
*/

#include "network/UDPSocket.hpp"
#include "boost/asio/ip/address.hpp"
#include "boost/asio/ip/udp.hpp"
#include <array>
#include <iostream>
#include <boost/bind.hpp>
#include <string>

using namespace network;

UDPSocket::UDPSocket(boost::asio::io_context &ioContext, unsigned short port):
    _socket(ioContext, boost::asio::ip::udp::endpoint(boost::asio::ip::udp::v4(), port))
{}

std::shared_ptr<UDPSocket> UDPSocket::create(boost::asio::io_context &ioContext, unsigned short port)
{
    return std::shared_ptr<UDPSocket>(new UDPSocket(ioContext, port));
}

boost::asio::ip::udp::socket &UDPSocket::socket()
{
    return _socket;
}

void UDPSocket::setReadFct(const std::function<int (const std::string &)> &handler)
{
    _readFct = handler;
}

void UDPSocket::read()
{
    if (_readFct == nullptr)
        return;

    _socket.async_receive(
        boost::asio::buffer(_bufferRead, sizeof(_bufferRead)),
        boost::bind(
            &UDPSocket::readHandler,
            this,
            boost::asio::placeholders::error,
            boost::asio::placeholders::bytes_transferred
        )
    );
}

void UDPSocket::readHandler(const boost::system::error_code &error, size_t size)
{
    if (size == 0 || error) {
        return handleDisconnect();
    }

    _savedBuffer.append(_bufferRead, size);
    while (_savedBuffer.size() != 0) {
        int toRemove = _readFct(_savedBuffer);

        if (toRemove == 0)
            break;

        _savedBuffer.erase(_savedBuffer.begin(), _savedBuffer.begin() + toRemove);
    }
    read();
}

void UDPSocket::setWriteFct(const std::function<void ()> &handler)
{
    _writeFct = handler;
}

void UDPSocket::write(const std::string &toWrite)
{
    _socket.async_send(
        boost::asio::buffer(toWrite),
        [this] (const boost::system::error_code &error, size_t /* size */) -> void {
            if (error)
                return handleDisconnect();
            if (this->_writeFct != nullptr)
                this->_writeFct();
        }
    );
}

void UDPSocket::write_to(const std::string &toWrite, const std::string &ip, unsigned short port)
{
    boost::asio::ip::udp::endpoint endpoint(boost::asio::ip::make_address(ip), port);

    _socket.async_send_to(
        boost::asio::buffer(toWrite),
        endpoint,
        [this] (const boost::system::error_code &error, size_t /* size */) -> void {
            if (error)
                return handleDisconnect();
            if (this->_writeFct != nullptr)
                this->_writeFct();
        }
    );
}

void UDPSocket::write(const std::string &toWrite, int /* size */)
{
    this->write(toWrite);
}

bool UDPSocket::isConnected() const
{
    return _connected;
}

void UDPSocket::setDisconnectFct(const std::function<void ()> &fct)
{
    _disconnectFct = fct;
}

void UDPSocket::handleDisconnect()
{
    _connected = false;
    if (_disconnectFct)
        _disconnectFct();
}

std::array<unsigned char, 4> UDPSocket::getIp() const
{
    std::cout << _socket.remote_endpoint().address().to_v4() << std::endl;
    return _socket.remote_endpoint().address().to_v4().to_bytes();
}

void UDPSocket::asyncConnect(
    const std::string &ip,
    unsigned short port,
    const std::function<void(const boost::system::error_code& ec)> &handler
)
{
    auto endpoint = boost::asio::ip::udp::endpoint(boost::asio::ip::make_address(ip), port);

    _socket.async_connect(endpoint, handler);
}

void UDPSocket::asyncConnect(
    const std::array<unsigned char, 4> &ip,
    unsigned short port,
    const std::function<void(const boost::system::error_code& error)> &fct
)
{
    std::stringstream sstream;

    sstream << ip[0] << "." << ip[1] << "." << ip[2] << "." << ip[3];
    asyncConnect(sstream.str(), port, fct);
}
